/*--------------------------------------------------------------
 *
 *	xneko
 *
 *			Original Writer:
 *            Masayuki Koba
 *
 *			Programmed by:
 *            Masayuki Koba, 1990
 *
 *          Added Color, RootWindow Capability and Quit Ability:
 *            Dan Checkoway, 7-12-94
 *
 *--------------------------------------------------------------*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <stdio.h>

#include <signal.h>
#include <math.h>
#include <sys/time.h>

#include "bitmaps/icon.xbm"
#include "bitmaps/cursor.xbm"
#include "bitmaps/cursor_mask.xbm"

#include "bitmaps/space.xbm"

#include "bitmaps/mati2.xbm"
#include "bitmaps/jare2.xbm"
#include "bitmaps/kaki1.xbm"
#include "bitmaps/kaki2.xbm"
#include "bitmaps/mati3.xbm"
#include "bitmaps/sleep1.xbm"
#include "bitmaps/sleep2.xbm"

#include "bitmaps/awake.xbm"

#include "bitmaps/up1.xbm"
#include "bitmaps/up2.xbm"
#include "bitmaps/down1.xbm"
#include "bitmaps/down2.xbm"
#include "bitmaps/left1.xbm"
#include "bitmaps/left2.xbm"
#include "bitmaps/right1.xbm"
#include "bitmaps/right2.xbm"
#include "bitmaps/upright1.xbm"
#include "bitmaps/upright2.xbm"
#include "bitmaps/upleft1.xbm"
#include "bitmaps/upleft2.xbm"
#include "bitmaps/dwleft1.xbm"
#include "bitmaps/dwleft2.xbm"
#include "bitmaps/dwright1.xbm"
#include "bitmaps/dwright2.xbm"

#include "bitmaps/utogi1.xbm"
#include "bitmaps/utogi2.xbm"
#include "bitmaps/dtogi1.xbm"
#include "bitmaps/dtogi2.xbm"
#include "bitmaps/ltogi1.xbm"
#include "bitmaps/ltogi2.xbm"
#include "bitmaps/rtogi1.xbm"
#include "bitmaps/rtogi2.xbm"

#define	BITMAP_WIDTH		32
#define	BITMAP_HEIGHT		32
#define	WINDOW_WIDTH		320
#define	WINDOW_HEIGHT		256
#define	DEFAULT_BORDER		2
#define	DEFAULT_WIN_X		1
#define	DEFAULT_WIN_Y		1
#define	AVAIL_KEYBUF		255

#define	EVENT_MASK       ( KeyPressMask | ButtonPressMask | \
						  ExposureMask | StructureNotifyMask )

#define	EVENT_MASK_ROOT  ( KeyPressMask | ExposureMask )

#define	MAX_TICK		9999		/* Odd Only! */
#define	INTERVAL		125000L
#define	NEKO_SPEED		16
#define	IDLE_SPACE		6
#define	NORMAL_STATE	1
#define	DEBUG_LIST		2
#define	DEBUG_MOVE		3

#define	NEKO_STOP		0
#define	NEKO_JARE		1
#define	NEKO_KAKI		2
#define	NEKO_AKUBI		3
#define	NEKO_SLEEP		4
#define	NEKO_AWAKE		5
#define	NEKO_U_MOVE		6
#define	NEKO_D_MOVE		7
#define	NEKO_L_MOVE		8
#define	NEKO_R_MOVE		9
#define	NEKO_UL_MOVE	10
#define	NEKO_UR_MOVE	11
#define	NEKO_DL_MOVE	12
#define	NEKO_DR_MOVE	13
#define	NEKO_U_TOGI		14
#define	NEKO_D_TOGI		15
#define	NEKO_L_TOGI		16
#define	NEKO_R_TOGI		17

#define	NEKO_STOP_TIME		4
#define	NEKO_JARE_TIME		10
#define	NEKO_KAKI_TIME		4
#define	NEKO_AKUBI_TIME		3
#define	NEKO_AWAKE_TIME		3
#define	NEKO_TOGI_TIME		10

#define	PI_PER8			((double)3.1415926535/(double)8)

#define	DIRNAMELEN		255

static int  useRoot;
char        *fgColor, *bgColor;

static char  *ProgramName;

Display        *theDisplay;
int            theScreen;
unsigned int   theDepth;
unsigned long  theBlackPixel;
unsigned long  theWhitePixel;
Window         theWindow;
Cursor         theCursor;

static unsigned int  WindowWidth;
static unsigned int  WindowHeight;
static int           WindowPointX;
static int           WindowPointY;

static unsigned int	BorderWidth = DEFAULT_BORDER;

long	IntervalTime = INTERVAL;

int		EventState;

int		NekoTickCount;
int		NekoStateCount;
int		NekoState;

int		MouseX;
int		MouseY;

int		PrevMouseX = 0;
int		PrevMouseY = 0;

int		NekoX;
int		NekoY;

int		NekoMoveDx;
int		NekoMoveDy;

int		NekoLastX;
int		NekoLastY;
GC		NekoLastGC;

double	NekoSpeed = (double)NEKO_SPEED;

double	SinPiPer8Times3;
double	SinPiPer8;

Pixmap	SpaceXbm;

Pixmap	Mati2Xbm;
Pixmap	Jare2Xbm;
Pixmap	Kaki1Xbm;
Pixmap	Kaki2Xbm;
Pixmap	Mati3Xbm;
Pixmap	Sleep1Xbm;
Pixmap	Sleep2Xbm;

Pixmap	AwakeXbm;

Pixmap	Up1Xbm;
Pixmap	Up2Xbm;
Pixmap	Down1Xbm;
Pixmap	Down2Xbm;
Pixmap	Left1Xbm;
Pixmap	Left2Xbm;
Pixmap	Right1Xbm;
Pixmap	Right2Xbm;
Pixmap	UpLeft1Xbm;
Pixmap	UpLeft2Xbm;
Pixmap	UpRight1Xbm;
Pixmap	UpRight2Xbm;
Pixmap	DownLeft1Xbm;
Pixmap	DownLeft2Xbm;
Pixmap	DownRight1Xbm;
Pixmap	DownRight2Xbm;

Pixmap	UpTogi1Xbm;
Pixmap	UpTogi2Xbm;
Pixmap	DownTogi1Xbm;
Pixmap	DownTogi2Xbm;
Pixmap	LeftTogi1Xbm;
Pixmap	LeftTogi2Xbm;
Pixmap	RightTogi1Xbm;
Pixmap	RightTogi2Xbm;

GC	SpaceGC;

GC	Mati2GC;
GC	Jare2GC;
GC	Kaki1GC;
GC	Kaki2GC;
GC	Mati3GC;
GC	Sleep1GC;
GC	Sleep2GC;

GC	AwakeGC;

GC	Up1GC;
GC	Up2GC;
GC	Down1GC;
GC	Down2GC;
GC	Left1GC;
GC	Left2GC;
GC	Right1GC;
GC	Right2GC;
GC	UpLeft1GC;
GC	UpLeft2GC;
GC	UpRight1GC;
GC	UpRight2GC;
GC	DownLeft1GC;
GC	DownLeft2GC;
GC	DownRight1GC;
GC	DownRight2GC;

GC	UpTogi1GC;
GC	UpTogi2GC;
GC	DownTogi1GC;
GC	DownTogi2GC;
GC	LeftTogi1GC;
GC	LeftTogi2GC;
GC	RightTogi1GC;
GC	RightTogi2GC;

typedef struct {
  GC            *GCCreatePtr;
  Pixmap        *BitmapCreatePtr;
  char          *PixelPattern;
  unsigned int  PixelWidth;
  unsigned int  PixelHeight;
} BitmapGCData;

BitmapGCData	BitmapGCDataTable[] = {
  { &SpaceGC, &SpaceXbm, space_bits, space_width, space_height },
  { &Mati2GC, &Mati2Xbm, mati2_bits, mati2_width, mati2_height },
  { &Jare2GC, &Jare2Xbm, jare2_bits, jare2_width, jare2_height },
  { &Kaki1GC, &Kaki1Xbm, kaki1_bits, kaki1_width, kaki1_height },
  { &Kaki2GC, &Kaki2Xbm, kaki2_bits, kaki2_width, kaki2_height },
  { &Mati3GC, &Mati3Xbm, mati3_bits, mati3_width, mati3_height },
  { &Sleep1GC, &Sleep1Xbm, sleep1_bits, sleep1_width, sleep1_height },
  { &Sleep2GC, &Sleep2Xbm, sleep2_bits, sleep2_width, sleep2_height },
  { &AwakeGC, &AwakeXbm, awake_bits, awake_width, awake_height },
  { &Up1GC, &Up1Xbm, up1_bits, up1_width, up1_height },
  { &Up2GC, &Up2Xbm, up2_bits, up2_width, up2_height },
  { &Down1GC, &Down1Xbm, down1_bits, down1_width, down1_height },
  { &Down2GC, &Down2Xbm, down2_bits, down2_width, down2_height },
  { &Left1GC, &Left1Xbm, left1_bits, left1_width, left1_height },
  { &Left2GC, &Left2Xbm, left2_bits, left2_width, left2_height },
  { &Right1GC, &Right1Xbm, right1_bits, right1_width, right1_height },
  { &Right2GC, &Right2Xbm, right2_bits, right2_width, right2_height },
  { &UpLeft1GC, &UpLeft1Xbm, upleft1_bits, upleft1_width, upleft1_height },
  { &UpLeft2GC, &UpLeft2Xbm, upleft2_bits, upleft2_width, upleft2_height },
  { &UpRight1GC,
	  &UpRight1Xbm, upright1_bits, upright1_width, upright1_height },
  { &UpRight2GC,
      &UpRight2Xbm, upright2_bits, upright2_width, upright2_height },
  { &DownLeft1GC, &DownLeft1Xbm, dwleft1_bits, dwleft1_width, dwleft1_height },
  { &DownLeft2GC, &DownLeft2Xbm, dwleft2_bits, dwleft2_width, dwleft2_height },
  { &DownRight1GC,
	  &DownRight1Xbm, dwright1_bits, dwright1_width, dwright1_height },
  { &DownRight2GC,
      &DownRight2Xbm, dwright2_bits, dwright2_width, dwright2_height },
  { &UpTogi1GC, &UpTogi1Xbm, utogi1_bits, utogi1_width, utogi1_height },
  { &UpTogi2GC, &UpTogi2Xbm, utogi2_bits, utogi2_width, utogi2_height },
  { &DownTogi1GC, &DownTogi1Xbm, dtogi1_bits, dtogi1_width, dtogi1_height },
  { &DownTogi2GC, &DownTogi2Xbm, dtogi2_bits, dtogi2_width, dtogi2_height },
  { &LeftTogi1GC, &LeftTogi1Xbm, ltogi1_bits, ltogi1_width, ltogi1_height },
  { &LeftTogi2GC, &LeftTogi2Xbm, ltogi2_bits, ltogi2_width, ltogi2_height },
  { &RightTogi1GC,
      &RightTogi1Xbm, rtogi1_bits, rtogi1_width, rtogi1_height },
  { &RightTogi2GC,
      &RightTogi2Xbm, rtogi2_bits, rtogi2_width, rtogi2_height },
  { NULL, NULL, NULL, NULL, NULL }
};

typedef struct {
  GC  *TickEvenGCPtr;
  GC  *TickOddGCPtr;
} Animation;

Animation  AnimationPattern[] = {
  { &Mati2GC, &Mati2GC },		/* NekoState == NEKO_STOP */
  { &Jare2GC, &Mati2GC },		/* NekoState == NEKO_JARE */
  { &Kaki1GC, &Kaki2GC },		/* NekoState == NEKO_KAKI */
  { &Mati3GC, &Mati3GC },		/* NekoState == NEKO_AKUBI */
  { &Sleep1GC, &Sleep2GC },		/* NekoState == NEKO_SLEEP */
  { &AwakeGC, &AwakeGC },		/* NekoState == NEKO_AWAKE */
  { &Up1GC, &Up2GC }	,		/* NekoState == NEKO_U_MOVE */
  { &Down1GC, &Down2GC },		/* NekoState == NEKO_D_MOVE */
  { &Left1GC, &Left2GC },		/* NekoState == NEKO_L_MOVE */
  { &Right1GC, &Right2GC },		/* NekoState == NEKO_R_MOVE */
  { &UpLeft1GC, &UpLeft2GC },		/* NekoState == NEKO_UL_MOVE */
  { &UpRight1GC, &UpRight2GC },	    /* NekoState == NEKO_UR_MOVE */
  { &DownLeft1GC, &DownLeft2GC },	/* NekoState == NEKO_DL_MOVE */
  { &DownRight1GC, &DownRight2GC },	/* NekoState == NEKO_DR_MOVE */
  { &UpTogi1GC, &UpTogi2GC },		/* NekoState == NEKO_U_TOGI */
  { &DownTogi1GC, &DownTogi2GC },	/* NekoState == NEKO_D_TOGI */
  { &LeftTogi1GC, &LeftTogi2GC },	/* NekoState == NEKO_L_TOGI */
  { &RightTogi1GC, &RightTogi2GC },	/* NekoState == NEKO_R_TOGI */
};


void  InitBitmapAndGCs() {
  BitmapGCData	*BitmapGCDataTablePtr;
  XGCValues		theGCValues;
  
  theGCValues.function = GXcopy;
  
  theGCValues.foreground = theBlackPixel;
  theGCValues.background = theWhitePixel;
  
  theGCValues.fill_style = FillTiled;
  
  for ( BitmapGCDataTablePtr = BitmapGCDataTable;
	   BitmapGCDataTablePtr->GCCreatePtr != NULL;
	   BitmapGCDataTablePtr++ ) {
	
	*(BitmapGCDataTablePtr->BitmapCreatePtr)
	  = XCreatePixmapFromBitmapData(
									theDisplay,
									RootWindow( theDisplay, theScreen ),
									BitmapGCDataTablePtr->PixelPattern,
									BitmapGCDataTablePtr->PixelWidth,
									BitmapGCDataTablePtr->PixelHeight,
									theBlackPixel, theWhitePixel,
									DefaultDepth( theDisplay, theScreen ) );
	
	theGCValues.tile = *(BitmapGCDataTablePtr->BitmapCreatePtr);
	
	*(BitmapGCDataTablePtr->GCCreatePtr)
	  = XCreateGC( theDisplay, theWindow,
				  GCFunction | GCForeground | GCBackground |
				  GCTile | GCFillStyle,
				  &theGCValues );
  }
  
  XFlush( theDisplay );
}


void
InitScreen( DisplayName, theGeometry, TitleName, iconicState )
char	*DisplayName;
char	*theGeometry;
char	*TitleName;
Bool	iconicState;
{
  int			GeometryStatus, status;
  Pixmap		theCursorSource;
  Pixmap		theCursorMask;
  Window		theRoot;
  Colormap		theColormap;
  XColor		theWhiteColor, theBlackColor, theExactColor;
  
  if ( ( theDisplay = XOpenDisplay( DisplayName ) ) == NULL ) {
	fprintf( stderr, "%s: Can't open display", ProgramName );
	if ( DisplayName != NULL )
	  fprintf( stderr, " %s.\n", DisplayName );
	else
	  fprintf( stderr, ".\n" );
	exit( 1 );
  }
  
  theScreen   = DefaultScreen( theDisplay );
  theDepth    = DefaultDepth( theDisplay, theScreen );
  theColormap = DefaultColormap( theDisplay, theScreen );
  
  GeometryStatus = XParseGeometry( theGeometry,
								  &WindowPointX, &WindowPointY,
								  &WindowWidth, &WindowHeight );
  
  if ( !( GeometryStatus & XValue ) )
	WindowPointX = DEFAULT_WIN_X;
  if ( !( GeometryStatus & YValue ) )
	WindowPointY = DEFAULT_WIN_Y;
  if ( !( GeometryStatus & WidthValue ) )
	WindowWidth = WINDOW_WIDTH;
  if ( !( GeometryStatus & HeightValue ) )
	WindowHeight = WINDOW_HEIGHT;
  
  theCursorSource = XCreateBitmapFromData( theDisplay,
										  RootWindow(theDisplay, theScreen),
										  cursor_bits,
										  cursor_width,
										  cursor_height );
  
  theCursorMask = XCreateBitmapFromData( theDisplay,
										RootWindow(theDisplay, theScreen),
										cursor_mask_bits,
										cursor_mask_width,
										cursor_mask_height );

  if ( bgColor )
	status = XAllocNamedColor( theDisplay, theColormap, bgColor,
							  &theWhiteColor, &theExactColor );
  else
	status = XAllocNamedColor( theDisplay, theColormap, "white",
							  &theWhiteColor, &theExactColor );
  
  if ( !status ) {
	fprintf( stderr,
			"%s: Can't allocate the background color.\n", ProgramName );
	exit( 1 );
  }
  
  if ( fgColor )
	status = XAllocNamedColor( theDisplay, theColormap, fgColor,
							  &theBlackColor, &theExactColor );
  else
	status = XAllocNamedColor( theDisplay, theColormap, "black",
							  &theBlackColor, &theExactColor );
  
  if ( !status ) {
	fprintf( stderr,
			"%s: Can't allocate the foreground color.\n", ProgramName );
	exit( 1 );
  }
  
  theBlackPixel = theBlackColor.pixel;
  theWhitePixel = theWhiteColor.pixel;
  
  theCursor = XCreatePixmapCursor( theDisplay,
								  theCursorSource, theCursorMask,
								  &theBlackColor, &theWhiteColor,
								  cursor_x_hot, cursor_y_hot );
  
  if ( useRoot ) {
	theWindow = RootWindow( theDisplay, theScreen );
	XDefineCursor( theDisplay, theWindow, theCursor );
	XSelectInput( theDisplay, theWindow, EVENT_MASK_ROOT );
  }
  else {
	XSetWindowAttributes  theWindowAttributes;
	XSizeHints            theSizeHints;
	XWMHints              theWMHints;
	unsigned long         theWindowMask;
	Pixmap                theIconPixmap;
	
	theWindowAttributes.border_pixel      = theBlackPixel;
	theWindowAttributes.background_pixel  = theWhitePixel;
	theWindowAttributes.cursor            = theCursor;
	theWindowAttributes.override_redirect = False;
	
	theWindowMask = CWBackPixel | CWBorderPixel |
	  CWCursor | CWOverrideRedirect;
	
	theWindow = XCreateWindow( theDisplay,
							  RootWindow( theDisplay, theScreen ),
							  WindowPointX, WindowPointY,
							  WindowWidth, WindowHeight,
							  BorderWidth, theDepth,
							  InputOutput, CopyFromParent,
							  theWindowMask, &theWindowAttributes );
	
	theIconPixmap = XCreateBitmapFromData( theDisplay, theWindow,
										  icon_bits, icon_width, icon_height );
	
	theWMHints.icon_pixmap = theIconPixmap;
	
	if ( iconicState )
	  theWMHints.initial_state = IconicState;
	else
	  theWMHints.initial_state = NormalState;
	
	theWMHints.flags = IconPixmapHint | StateHint;
	
	XSetWMHints( theDisplay, theWindow, &theWMHints );
	
	theSizeHints.flags = PPosition | PSize;
	theSizeHints.x = WindowPointX;
	theSizeHints.y = WindowPointY;
	theSizeHints.width = WindowWidth;
	theSizeHints.height = WindowHeight;
	
	XSetNormalHints( theDisplay, theWindow, &theSizeHints );
	
	if ( strlen( TitleName ) >= 1 ) {
	  XStoreName( theDisplay, theWindow, TitleName );
	  XSetIconName( theDisplay, theWindow, TitleName );
	} else {
	  XStoreName( theDisplay, theWindow, ProgramName );
	  XSetIconName( theDisplay, theWindow, ProgramName );
	}
	
	XMapWindow( theDisplay, theWindow );
	XSelectInput( theDisplay, theWindow, EVENT_MASK );
  }
  
  XSetWindowBackground( theDisplay, theWindow, theWhitePixel );
  XClearWindow( theDisplay, theWindow );
  XFlush( theDisplay );
  
  XGetGeometry( theDisplay, theWindow, &theRoot,
			   &WindowPointX, &WindowPointY, &WindowWidth, &WindowHeight,
			   &BorderWidth, &theDepth );
  
  InitBitmapAndGCs();
  
  XFlush( theDisplay );
}


void  Interval() {
  pause();
}


void  TickCount() {
  if ( ++NekoTickCount >= MAX_TICK )
	NekoTickCount = 0;
  
  if ( NekoTickCount % 2 == 0 )
	if ( NekoStateCount < MAX_TICK )
	  NekoStateCount++;
}


void
SetNekoState( SetValue )
int		SetValue;
{
  NekoTickCount = 0;
  NekoStateCount = 0;
  
  NekoState = SetValue;
  
#ifdef	DEBUG
  switch ( NekoState ) {
  case NEKO_STOP:
  case NEKO_JARE:
  case NEKO_KAKI:
  case NEKO_AKUBI:
  case NEKO_SLEEP:
  case NEKO_U_TOGI:
  case NEKO_D_TOGI:
  case NEKO_L_TOGI:
  case NEKO_R_TOGI:
	NekoMoveDx = NekoMoveDy = 0;
	break;
  default:
	break;
  }
#endif
}


void
DrawNeko( x, y, DrawGC )
int		x;
int		y;
GC		DrawGC;
{
  if ( (x != NekoLastX || y != NekoLastY) && (EventState != DEBUG_LIST) )
	XFillRectangle( theDisplay, theWindow, SpaceGC, NekoLastX, NekoLastY,
				   BITMAP_WIDTH, BITMAP_HEIGHT );
  
  XSetTSOrigin( theDisplay, DrawGC, x, y );
  
  XFillRectangle( theDisplay, theWindow, DrawGC,
				 x, y, BITMAP_WIDTH, BITMAP_HEIGHT );
  
  XFlush( theDisplay );
  
  NekoLastX = x;
  NekoLastY = y;
  
  NekoLastGC = DrawGC;
}


void  RedrawNeko() {
  XFillRectangle( theDisplay, theWindow, NekoLastGC, NekoLastX, NekoLastY,
				 BITMAP_WIDTH, BITMAP_HEIGHT );
  XFlush( theDisplay );
}


void  NekoDirection() {
  int			NewState;
  double		LargeX, LargeY;
  double		Length;
  double		SinTheta;
  
  if ( NekoMoveDx == 0 && NekoMoveDy == 0 ) {
	NewState = NEKO_STOP;
  } else {
	LargeX = (double)NekoMoveDx;
	LargeY = (double)(-NekoMoveDy);
	Length = sqrt( LargeX * LargeX + LargeY * LargeY );
	SinTheta = LargeY / Length;
	
	if ( NekoMoveDx > 0 ) {
	  if ( SinTheta > SinPiPer8Times3 ) {
		NewState = NEKO_U_MOVE;
	  } else if ( ( SinTheta <= SinPiPer8Times3 )
				 && ( SinTheta > SinPiPer8 ) ) {
		NewState = NEKO_UR_MOVE;
	  } else if ( ( SinTheta <= SinPiPer8 )
				 && ( SinTheta > -( SinPiPer8 ) ) ) {
		NewState = NEKO_R_MOVE;
	  } else if ( ( SinTheta <= -( SinPiPer8 ) )
				 && ( SinTheta > -( SinPiPer8Times3 ) ) ) {
		NewState = NEKO_DR_MOVE;
	  } else {
		NewState = NEKO_D_MOVE;
	  }
	} else {
	  if ( SinTheta > SinPiPer8Times3 ) {
		NewState = NEKO_U_MOVE;
	  } else if ( ( SinTheta <= SinPiPer8Times3 )
				 && ( SinTheta > SinPiPer8 ) ) {
		NewState = NEKO_UL_MOVE;
	  } else if ( ( SinTheta <= SinPiPer8 )
				 && ( SinTheta > -( SinPiPer8 ) ) ) {
		NewState = NEKO_L_MOVE;
	  } else if ( ( SinTheta <= -( SinPiPer8 ) )
				 && ( SinTheta > -( SinPiPer8Times3 ) ) ) {
		NewState = NEKO_DL_MOVE;
	  } else {
		NewState = NEKO_D_MOVE;
	  }
	}
  }
  
  if ( NekoState != NewState ) {
	SetNekoState( NewState );
  }
}


Bool  IsWindowOver() {
  Bool	ReturnValue = False;
  
  if ( NekoY <= 0 ) {
	NekoY = 0;
	ReturnValue = True;
  } else if ( NekoY >= WindowHeight - BITMAP_HEIGHT ) {
	NekoY = WindowHeight - BITMAP_HEIGHT;
	ReturnValue = True;
  }
  if ( NekoX <= 0 ) {
	NekoX = 0;
	ReturnValue = True;
  } else if ( NekoX >= WindowWidth - BITMAP_WIDTH ) {
	NekoX = WindowWidth - BITMAP_WIDTH;
	ReturnValue = True;
  }
  
  return( ReturnValue );
}


Bool  IsNekoDontMove() {
  return( NekoX == NekoLastX  &&  NekoY == NekoLastY );
}


Bool  IsNekoMoveStart() {
#ifndef	DEBUG
  if ( (PrevMouseX >= MouseX - IDLE_SPACE
		&& PrevMouseX <= MouseX + IDLE_SPACE) &&
	  (PrevMouseY >= MouseY - IDLE_SPACE 
	   && PrevMouseY <= MouseY + IDLE_SPACE) )
	return( False );
  else
	return( True );
#else
  if ( NekoMoveDx == 0 && NekoMoveDy == 0 )
	return( False );
  else
	return( True );
#endif
}


void  CalcDxDy() {
  Window		QueryRoot, QueryChild;
  int			AbsoluteX, AbsoluteY;
  int			RelativeX, RelativeY;
  unsigned int	ModKeyMask;
  double		LargeX, LargeY;
  double		DoubleLength, Length;
  
  XQueryPointer( theDisplay, theWindow, &QueryRoot, &QueryChild,
				&AbsoluteX, &AbsoluteY, &RelativeX, &RelativeY, &ModKeyMask );
  
  PrevMouseX = MouseX;
  PrevMouseY = MouseY;
  
  MouseX = RelativeX;
  MouseY = RelativeY;
  
  LargeX = (double)( MouseX - NekoX - BITMAP_WIDTH / 2 );
  LargeY = (double)( MouseY - NekoY - BITMAP_HEIGHT );
  
  DoubleLength = LargeX * LargeX + LargeY * LargeY;
  
  if ( DoubleLength != (double)0 ) {
	Length = sqrt( DoubleLength );
	if ( Length <= NekoSpeed ) {
	  NekoMoveDx = (int)LargeX;
	  NekoMoveDy = (int)LargeY;
	} else {
	  NekoMoveDx = (int)( ( NekoSpeed * LargeX ) / Length );
	  NekoMoveDy = (int)( ( NekoSpeed * LargeY ) / Length );
	}
  } else {
	NekoMoveDx = NekoMoveDy = 0;
  }
}


void  NekoThinkDraw() {
#ifndef	DEBUG
  CalcDxDy();
#endif

  if ( NekoState != NEKO_SLEEP ) {
	DrawNeko( NekoX, NekoY,
			 NekoTickCount % 2 == 0 ?
			 *(AnimationPattern[ NekoState ].TickEvenGCPtr) :
			 *(AnimationPattern[ NekoState ].TickOddGCPtr) );
  } else {
	DrawNeko( NekoX, NekoY,
			 NekoTickCount % 8 <= 3 ?
			 *(AnimationPattern[ NekoState ].TickEvenGCPtr) :
			 *(AnimationPattern[ NekoState ].TickOddGCPtr) );
  }
  
  TickCount();
  
  switch ( NekoState ) {
  case NEKO_STOP:
	if ( IsNekoMoveStart() ) {
	  SetNekoState( NEKO_AWAKE );
	  break;
	}
	if ( NekoStateCount < NEKO_STOP_TIME ) {
	  break;
	}
	if ( NekoMoveDx < 0 && NekoX <= 0 ) {
	  SetNekoState( NEKO_L_TOGI );
	} else if ( NekoMoveDx > 0 && NekoX >= WindowWidth - BITMAP_WIDTH ) {
	  SetNekoState( NEKO_R_TOGI );
	} else if ( NekoMoveDy < 0 && NekoY <= 0 ) {
	  SetNekoState( NEKO_U_TOGI );
	} else if ( NekoMoveDy > 0 && NekoY >= WindowHeight - BITMAP_HEIGHT ) {
	  SetNekoState( NEKO_D_TOGI );
	} else {
	  SetNekoState( NEKO_JARE );
	}
	break;
  case NEKO_JARE:
	if ( IsNekoMoveStart() ) {
	  SetNekoState( NEKO_AWAKE );
	  break;
	}
	if ( NekoStateCount < NEKO_JARE_TIME ) {
	  break;
	}
	SetNekoState( NEKO_KAKI );
	break;
  case NEKO_KAKI:
	if ( IsNekoMoveStart() ) {
	  SetNekoState( NEKO_AWAKE );
	  break;
	}
	if ( NekoStateCount < NEKO_KAKI_TIME ) {
	  break;
	}
	SetNekoState( NEKO_AKUBI );
	break;
  case NEKO_AKUBI:
	if ( IsNekoMoveStart() ) {
	  SetNekoState( NEKO_AWAKE );
	  break;
	}
	if ( NekoStateCount < NEKO_AKUBI_TIME ) {
	  break;
	}
	SetNekoState( NEKO_SLEEP );
	break;
  case NEKO_SLEEP:
	if ( IsNekoMoveStart() ) {
	  SetNekoState( NEKO_AWAKE );
	  break;
	}
	break;
  case NEKO_AWAKE:
	if ( NekoStateCount < NEKO_AWAKE_TIME ) {
	  break;
	}
	NekoDirection();
	break;
  case NEKO_U_MOVE:
  case NEKO_D_MOVE:
  case NEKO_L_MOVE:
  case NEKO_R_MOVE:
  case NEKO_UL_MOVE:
  case NEKO_UR_MOVE:
  case NEKO_DL_MOVE:
  case NEKO_DR_MOVE:
	NekoX += NekoMoveDx;
	NekoY += NekoMoveDy;
	NekoDirection();
	if ( IsWindowOver() ) {
	  if ( IsNekoDontMove() ) {
		SetNekoState( NEKO_STOP );
	  }
	}
	break;
  case NEKO_U_TOGI:
  case NEKO_D_TOGI:
  case NEKO_L_TOGI:
  case NEKO_R_TOGI:
	if ( IsNekoMoveStart() ) {
	  SetNekoState( NEKO_AWAKE );
	  break;
	}
	if ( NekoStateCount < NEKO_TOGI_TIME ) {
	  break;
	}
	SetNekoState( NEKO_KAKI );
	break;
  default:
	/* Internal Error */
	SetNekoState( NEKO_STOP );
	break;
  }
  
  Interval();
}


#ifdef	DEBUG
void  DisplayCharacters() {
  int		Index;
  int		x, y;
  
  for ( Index = 0, x = 0, y = 0;
	   BitmapGCDataTable[ Index ].GCCreatePtr != NULL; Index++ ) {
	
	DrawNeko( x, y, *(BitmapGCDataTable[ Index ].GCCreatePtr) );
	XFlush( theDisplay );
	
	x += BITMAP_WIDTH;
	
	if ( x > WindowWidth - BITMAP_WIDTH ) {
	  x = 0;
	  y += BITMAP_HEIGHT;
	  if ( y > WindowHeight - BITMAP_HEIGHT) {
		break;
	  }
	}
  }
}

#endif	/* DEBUG */


Bool
ProcessKeyPress( theKeyEvent )
XKeyEvent	*theKeyEvent;
{
  int			Length;
  int			theKeyBufferMaxLen = AVAIL_KEYBUF;
  char		theKeyBuffer[ AVAIL_KEYBUF + 1 ];
  KeySym		theKeySym;
  XComposeStatus	theComposeStatus;
  Bool		ReturnState;
  
  ReturnState = True;
  
  Length = XLookupString( theKeyEvent,
						 theKeyBuffer, theKeyBufferMaxLen,
						 &theKeySym, &theComposeStatus );
  
  if ( Length > 0 ) {
	switch ( theKeyBuffer[ 0 ] ) {
	case 'q':
	case 'Q':
	  if ( theKeyEvent->state & Mod1Mask ) {	/* META (Alt) %-!< */
		ReturnState = False;
	  }
	  break;
	default:
	  break;
	}
  }
  
#ifdef	DEBUG
  if ( EventState == DEBUG_MOVE ) {
	switch ( theKeySym ) {
	case XK_KP_1:
	  NekoMoveDx = -(int)( NekoSpeed / sqrt( (double)2 ) );
	  NekoMoveDy = -NekoMoveDx;
	  break;
	case XK_KP_2:
	  NekoMoveDx = 0;
	  NekoMoveDy = (int)NekoSpeed;
	  break;
	case XK_KP_3:
	  NekoMoveDx = (int)( NekoSpeed / sqrt( (double)2 ) );
	  NekoMoveDy = NekoMoveDx;
	  break;
	case XK_KP_4:
	  NekoMoveDx = -(int)NekoSpeed;
	  NekoMoveDy = 0;
	  break;
	case XK_KP_5:
	  NekoMoveDx = 0;
	  NekoMoveDy = 0;
	  break;
	case XK_KP_6:
	  NekoMoveDx = (int)NekoSpeed;
	  NekoMoveDy = 0;
	  break;
	case XK_KP_7:
	  NekoMoveDx = -(int)( NekoSpeed / sqrt( (double)2 ) );
	  NekoMoveDy = NekoMoveDx;
	  break;
	case XK_KP_8:
	  NekoMoveDx = 0;
	  NekoMoveDy = -(int)NekoSpeed;
	  break;
	case XK_KP_9:
	  NekoMoveDx = (int)( NekoSpeed / sqrt( (double)2 ) );
	  NekoMoveDy = -NekoMoveDx;
	  break;
	}
  }
#endif
  
  return( ReturnState );
}


void  NekoAdjust() {
  if ( NekoX < 0 )
	NekoX = 0;
  else if ( NekoX > WindowWidth - BITMAP_WIDTH )
	NekoX = WindowWidth - BITMAP_WIDTH;

  if ( NekoY < 0 )
	NekoY = 0;
  else if ( NekoY > WindowHeight - BITMAP_HEIGHT )
	NekoY = WindowHeight - BITMAP_HEIGHT;
}


Bool  ProcessEvent() {
  XEvent	theEvent;
  Bool	ContinueState = True;
  
  switch ( EventState ) {
  case NORMAL_STATE:
	while ( XCheckMaskEvent( theDisplay, EVENT_MASK, &theEvent ) ) {
	  switch ( theEvent.type ) {
	  case ConfigureNotify:
		WindowWidth = theEvent.xconfigure.width;
		WindowHeight = theEvent.xconfigure.height;
		WindowPointX = theEvent.xconfigure.x;
		WindowPointY = theEvent.xconfigure.y;
		BorderWidth = theEvent.xconfigure.border_width;
		NekoAdjust();
		break;
	  case Expose:
		if ( theEvent.xexpose.count == 0 )
		  RedrawNeko();
		break;
	  case MapNotify:
		RedrawNeko();
		break;
	  case KeyPress:
		ContinueState = ProcessKeyPress( &theEvent.xkey );
		if ( !ContinueState ) {
		  return( ContinueState );
		}
		break;
	  case ButtonPress:
		if ( theEvent.xbutton.button == 3 )
		  return( False );
		break;
	  default:
		/* Unknown Event */
		break;
	  }
	}
	break;
#ifdef	DEBUG
  case DEBUG_LIST:
	XNextEvent( theDisplay, &theEvent );
	switch ( theEvent.type ) {
	case ConfigureNotify:
	  WindowWidth = theEvent.xconfigure.width;
	  WindowHeight = theEvent.xconfigure.height;
	  WindowPointX = theEvent.xconfigure.x;
	  WindowPointY = theEvent.xconfigure.y;
	  BorderWidth = theEvent.xconfigure.border_width;
	  break;
	case Expose:
	  if ( theEvent.xexpose.count == 0 )
		DisplayCharacters();
	  break;
	case MapNotify:
	  DisplayCharacters();
	  break;
	case KeyPress:
	  ContinueState = ProcessKeyPress( &theEvent );
	  break;
	case ButtonPress:
	  if ( theEvent.xbutton.button == 3 )
		return( False );
	  break;
	default:
	  /* Unknown Event */
	  break;
	}
	break;
  case DEBUG_MOVE:
	while ( XCheckMaskEvent( theDisplay, EVENT_MASK, &theEvent ) ) {
	  switch ( theEvent.type ) {
	  case ConfigureNotify:
		WindowWidth = theEvent.xconfigure.width;
		WindowHeight = theEvent.xconfigure.height;
		WindowPointX = theEvent.xconfigure.x;
		WindowPointY = theEvent.xconfigure.y;
		BorderWidth = theEvent.xconfigure.border_width;
		NekoAdjust();
		break;
	  case Expose:
		if ( theEvent.xexpose.count == 0 )
		  RedrawNeko();
		break;
	  case MapNotify:
		RedrawNeko();
		break;
	  case KeyPress:
		ContinueState = ProcessKeyPress( &theEvent );
		if ( !ContinueState ) {
		  return( ContinueState );
		}
		break;
	  case ButtonPress:
		if ( theEvent.xbutton.button == 3 )
		  return( False );
		break;
	  default:
		/* Unknown Event */
		break;
	  }
	}
	break;
#endif
  default:
	/* Internal Error */
	break;
  }
  
  return( ContinueState );
}


void  ProcessNeko() {
  struct itimerval	Value;
  
  EventState = NORMAL_STATE;

  NekoX = ( WindowWidth - BITMAP_WIDTH / 2 ) / 2;
  NekoY = ( WindowHeight - BITMAP_HEIGHT / 2 ) / 2;
  
  NekoLastX = NekoX;
  NekoLastY = NekoY;
  
  SetNekoState( NEKO_STOP );
  
  timerclear( &Value.it_interval );
  timerclear( &Value.it_value );
  
  Value.it_interval.tv_usec = IntervalTime;
  Value.it_value.tv_usec = IntervalTime;
  
  setitimer( ITIMER_REAL, &Value, 0 );
  
  do {
	NekoThinkDraw();
  } while ( ProcessEvent() );
}


#ifdef	DEBUG

void  NekoList() {
  EventState = DEBUG_LIST;
  
  fprintf( stderr, "\n" );
  fprintf( stderr, "G-0lMw$rI=<($7$^$9!#(Quit !D Alt-Q)\n" );
  fprintf( stderr, "\n" );
  
  XSelectInput( theDisplay, theWindow, EVENT_MASK );
  
  while ( ProcessEvent() );
}


void  NekoMoveTest() {
  struct itimerval	Value;
  
  EventState = DEBUG_MOVE;
  
  NekoX = ( WindowWidth - BITMAP_WIDTH / 2 ) / 2;
  NekoY = ( WindowHeight - BITMAP_HEIGHT / 2 ) / 2;
  
  NekoLastX = NekoX;
  NekoLastY = NekoY;
  
  SetNekoState( NEKO_STOP );
  
  timerclear( &Value.it_interval );
  timerclear( &Value.it_value );
  
  Value.it_interval.tv_usec = IntervalTime;
  Value.it_value.tv_usec = IntervalTime;
  
  setitimer( ITIMER_REAL, &Value, 0 );
  
  fprintf( stderr, "\n" );
  fprintf( stderr, "G-$N0\F0%F%9%H$r9T$$$^$9!#(Quit !D Alt-Q)\n" );
  fprintf( stderr, "\n" );
  fprintf( stderr, "\t%-!<%Q%C%I>e$N%F%s%-!<$GG-$r0\F0$5$;$F2<$5$$!#\n" );
  fprintf( stderr, "\t(M-8z$J%-!<$O#1!A#9$G$9!#)\n" );
  fprintf( stderr, "\n" );
  
  do {
	NekoThinkDraw();
  } while ( ProcessEvent() );
}


void  ProcessDebugMenu() {
  int		UserSelectNo = 0;
  char	UserAnswer[ BUFSIZ ];
  
  fprintf( stderr, "\n" );
  fprintf( stderr, "!Zxneko %G%P%C%0%a%K%e!<![\n" );
  
  while ( !( UserSelectNo >= 1 && UserSelectNo <= 2 ) ) {
	fprintf( stderr, "\n" );
	fprintf( stderr, "\t1)!!G-%-%c%i%/%?!<0lMwI=<(\n" );
	fprintf( stderr, "\t2)!!G-0\F0%F%9%H\n" );
	fprintf( stderr, "\n" );
	fprintf( stderr, "Select: " );

	fgets( UserAnswer, sizeof( UserAnswer ), stdin );
	
	UserSelectNo = atoi( UserAnswer );
	
	if ( !( UserSelectNo >= 1 && UserSelectNo <= 2 ) ) {
	  fprintf( stderr, "\n" );
	  fprintf( stderr, "@5$7$$HV9f$rA*Br$7$F2<$5$$!#\n" );
	}
  }
  
  switch ( UserSelectNo ) {
  case 1:
	NekoList();
	break;
  case 2:
	NekoMoveTest();
	break;
  default:
	/* Internal Error */
	break;
  }
  
  fprintf( stderr, "%F%9%H=*N;!#\n" );
  fprintf( stderr, "\n" );
}

#endif	/* DEBUG */


void  NullFunction()
{
  signal( SIGALRM, NullFunction );
}


void  Usage() {
  fprintf( stderr,
		  "Usage: %s [-display <display>] [-geometry <geometry>] \\\n",
		  ProgramName );
  fprintf( stderr, "\t[-bg <background>] [-fg <foreground>] \\\n" );
  fprintf( stderr, "\t[-title <title>] [-name <title>] [-iconic] \\\n" );
  fprintf( stderr, "\t[-speed <speed>] [-time <time>] [-root] [-help]\n" );
}


Bool
GetArguments( argc, argv, theDisplayName, theGeometry, theTitle,
	      NekoSpeed, IntervalTime )
int		argc;
char	*argv[];
char	*theDisplayName;
char	*theGeometry;
char	*theTitle;
double	*NekoSpeed;
long	*IntervalTime;
{
  int		ArgCounter;
  Bool	iconicState;
  
  theDisplayName[ 0 ] = '\0';
  theGeometry[ 0 ] = '\0';
  theTitle[ 0 ] = '\0';
  
  iconicState = False;
  
  for ( ArgCounter = 0; ArgCounter < argc; ArgCounter++ ) {
	if ( strncmp( argv[ ArgCounter ], "-h", 2 ) == 0 ) {
	  Usage();
	  exit( 0 );
	} else if ( strcmp( argv[ ArgCounter ], "-display" ) == 0 ) {
	  ArgCounter++;
	  if ( ArgCounter < argc ) {
		strcpy( theDisplayName, argv[ ArgCounter ] );
	  } else {
		fprintf( stderr, "%s: -display option error.\n", ProgramName );
		exit( 1 );
	  }
	} else if ( strncmp( argv[ ArgCounter ], "-geom", 5 ) == 0 ) {
	  ArgCounter++;
	  if ( ArgCounter < argc ) {
		strcpy( theGeometry, argv[ ArgCounter ] );
	  } else {
		fprintf( stderr,
				"%s: -geometry option error.\n", ProgramName );
		exit( 1 );
	  }
	} else if ( ( strcmp( argv[ ArgCounter ], "-title" ) == 0 )
			   || ( strcmp( argv[ ArgCounter ], "-name" ) == 0 ) ) {
	  ArgCounter++;
	  if ( ArgCounter < argc ) {
		strcpy( theTitle, argv[ ArgCounter ] );
	  } else {
		fprintf( stderr, "%s: -title option error.\n", ProgramName );
		exit( 1 );
	  }
	} else if ( strcmp( argv[ ArgCounter ], "-iconic" ) == 0 ) {
	  iconicState = True;
	} else if ( strcmp( argv[ ArgCounter ], "-speed" ) == 0 ) {
	  ArgCounter++;
	  if ( ArgCounter < argc ) {
		*NekoSpeed = atof( argv[ ArgCounter ] );
	  } else {
		fprintf( stderr, "%s: -speed option error.\n", ProgramName );
		exit( 1 );
	  }
	} else if ( strcmp( argv[ ArgCounter ], "-fg" ) == 0 ) {
	  ArgCounter++;
	  if ( ArgCounter < argc ) {
		fgColor = argv[ArgCounter];
	  } else {
		fprintf( stderr, "%s: -fg option error.\n", ProgramName );
		exit( 1 );
	  }
	} else if ( strcmp( argv[ ArgCounter ], "-bg" ) == 0 ) {
	  ArgCounter++;
	  if ( ArgCounter < argc ) {
		bgColor = argv[ArgCounter];
	  } else {
		fprintf( stderr, "%s: -bg option error.\n", ProgramName );
		exit( 1 );
	  }
	} else if ( strcmp( argv[ ArgCounter ], "-time" ) == 0 ) {
	  ArgCounter++;
	  if ( ArgCounter < argc ) {
		*IntervalTime = atol( argv[ ArgCounter ] );
	  } else {
		fprintf( stderr, "%s: -time option error.\n", ProgramName );
		exit( 1 );
	  }
	} else if ( strcmp( argv[ ArgCounter ], "-root" ) == 0 ) {
	  useRoot = 1;
	} else {
	  fprintf( stderr,
			  "%s: Unknown option \"%s\".\n", ProgramName,
			  argv[ ArgCounter ] );
	  Usage();
	  exit( 1 );
	}
  }
  
  if ( strlen( theDisplayName ) < 1 )
	theDisplayName = NULL;
  
  if ( strlen( theGeometry ) < 1 )
	theGeometry = NULL;
  
  return( iconicState );
}


int
main( argc, argv )
int		argc;
char	*argv[];
{
  Bool	iconicState;
  char	theDisplayName[ DIRNAMELEN ];
  char	theGeometry[ DIRNAMELEN ];
  char	theTitle[ DIRNAMELEN ];
  
  ProgramName = argv[ 0 ];
  
  argc--;
  argv++;
  
  useRoot = 0;
  fgColor = bgColor = (char *) NULL;
  
  iconicState = GetArguments( argc, argv, theDisplayName, theGeometry,
							 theTitle, &NekoSpeed, &IntervalTime );
  
  InitScreen( theDisplayName, theGeometry, theTitle, iconicState );
  
  signal( SIGALRM, NullFunction );
  
  SinPiPer8Times3 = sin( PI_PER8 * (double)3 );
  SinPiPer8 = sin( PI_PER8 );
  
#ifndef	DEBUG
  ProcessNeko();
#else
  ProcessDebugMenu();
#endif
  
  XUndefineCursor( theDisplay, theWindow );
  exit( 0 );
}
